# -*- coding: utf-8 -*-
"""
Created on Wed Oct 12 15:34:02 2016

@author: rmcleod
"""
import blosc
import numpy as np
import time
import mrcz
import tempfile, os, os.path, glob

print( "==== Generating simulated Poisson stacks ====" )
print( "Note: will write files to: %s. (they will not be deleted as they can be re-used!)" % tempfile.gettempdir() )


stackShape = [50,4096,4096]
doseLevelList = [ 0.01, 0.1, 0.25, 0.5, 1.0, 1.5, 2.0, 4.0 ]
doseFilenames = []
poissonDataArray = np.zeros( [len(doseLevelList), stackShape[0],stackShape[1],stackShape[2]]  )
for I, dose in enumerate(doseLevelList):
    poissonName = os.path.join(tempfile.gettempdir(), "poisson%f.mrcz" % dose )
    doseFilenames.append( poissonName )
 
    if not os.path.isfile( poissonName ):
        print( "Generating Poisson random numbers and writing %s" % poissonName )
        poissonDataArray[I,...] = np.random.poisson( lam=dose, size=stackShape ).astype('int8')
        mrcz.MRCExport( poissonDataArray[I,...], poissonName, compressor='zstd' )

    else:
        # Load from disk
        print( "Loading from disk: %s" % poissonName )
        poissonDataArray[I,...] = mrcz.MRCImport( poissonName )
        

print( "==== Starting compression benchmark ====" )

def doCompression( dataStack, 
                  compressor='zstd', blocksize=2**20, n_threads=16, 
                  shuffle=blosc.BITSHUFFLE, clevel=5 ):

    blosc.set_blocksize( blocksize )
    blosc.set_nthreads( n_threads )
    typeSize = dataStack.dtype.itemsize
    packedDataList = [None] * dataStack.shape[0]
    for J in np.arange(dataStack.shape[0]):
        packedDataList[J] = blosc.compress( dataStack[J,:,:], typesize=typeSize,
                                     clevel=clevel, shuffle=shuffle, cname=compressor )
        
    return packedDataList

def doDecompression( packedDataList, shape, n_threads ):
    blosc.set_nthreads( n_threads )
    dataList = [None] * len(packedDataList)
    for J in np.arange(len(packedDataList) ):
#        dataStack[J,:,:] = np.reshape( 
#            np.frombuffer( blosc.decompress( packedDataList[J] ), dtype='uint8' ),
#            shape[1:] )
        # Something here Numpy-side is very slow, so let's not include that in our 
        # benchmark.
        dataList[J] = blosc.decompress( packedDataList[J] )
    return dataList




#t_half0 = time.time()
#halfimage = dm4image_8bit[:,:,::2] + np.left_shift(dm4image_8bit[:,:,1::2],4)
#t_half1 = time.time()
#restoreimage = np.empty( header['dimensions'], dtype='uint8' )
##image[0::2] = np.left_shift(interlaced_image,4)/16
##image[1::2] = np.right_shift(interlaced_image,4)
## Different interlace option
## TODO: AND array with 15 instead?
#restoreimage[:,:,::2] = (np.left_shift( halfimage, 4 ) & 15 )
#restoreimage[:,:,1::2] = np.right_shift( halfimage, 4 )
#t_half2 = time.time()
#
#print( "4-byte encoding time (s): %f" % (t_half1 - t_half0) )
#print( "4-byte DEcoding time (s): %f" % (t_half2 - t_half1) )

originalBytes = 50*4096*4096*4 # Guess of size if we saved as float-32


# Something is wrong with lz4hc, it's really not competitive
# All codecs slow down if you use bloscLZ for some reason. blocksize adjustment?

# Because our data is bit-limited by shot noise generally we will always 
# want to use the BITSHUFFLE filter when working with a counting electron detector
# Probably BITSHUFFLE is not so great for floating-point data, however.
SHUFFLE = blosc.BITSHUFFLE
nThreadsList = np.arange( 12, 48+1, 6 )
compressorList = np.array( ['lz4', 'zlib', 'zstd' ] )
clevelList = np.arange(1,7+1)

testRuns = 1
blockSizeList = np.array( [2**15, 2**16, 2**17, 2**18, 2**19, 2**20, 2**21, 2**22], dtype='int64' )

optiShape = [len(doseLevelList), len(compressorList), len(clevelList), 
             len(nThreadsList), len(blockSizeList), testRuns]
cCompress = np.zeros( optiShape )
cBytes = np.zeros( optiShape )
cRatio = np.zeros( optiShape )
cDecompress = np.zeros( optiShape )



for I, doseLevel in enumerate( doseLevelList ):
    for J, compressor in enumerate(compressorList):
        for K, clevel in enumerate( clevelList ):
            for N, N_THREADS in enumerate( nThreadsList ):
                for B, blocksize in enumerate( blockSizeList ):
                    
                    print( "Testing compressor %s level %d with %d threads and blocksize %d at dose level %f counts/pix" % 
                        (compressor, clevel, N_THREADS, blocksize, doseLevel) )
                            
                    for L in np.arange(testRuns):
    
                        t0 = time.time()
                        packedData = doCompression( poissonDataArray[I, ...], 
                                                       compressor=compressor,
                                                       n_threads=N_THREADS, 
                                                       shuffle=SHUFFLE, 
                                                       clevel=clevel,
                                                       blocksize=blocksize )
                        t1 = time.time()
                        cCompress[I,J,K,N,B,L] = t1 - t0
                        for index in np.arange( len(packedData) ):
                            cBytes[I,J,K,N,B,L] += len( packedData[index] )
                        cRatio[I,J,K,N,B,L] = 100.0 * originalBytes / cBytes[I,J,K,N,B,L]
                        
                        t2 = time.time()
                        # Will this work on all frames?  No....
                        result = doDecompression( packedData, len(packedData), N_THREADS )
                        t3 = time.time()
                        cDecompress[I,J,K,N,B,L] = t3 - t2

np.save( "cCompress.npy", cCompress )
np.save( "cBytes.npy", cBytes )
np.save( "cRatio.npy", cRatio )
np.save( "cDecompress.npy", cDecompress )

#print( "4-byte encoding time (s): %f" % (t_half1 - t_half0) )
#print( "4-byte DEncoding time (s): %f" % (t_half2 - t_half1) )
print( "==== Compressor ====" )
print( compressorList )
print( "==== Compression Time (s) ====" )
print( cCompress )
print( "==== DEcompression Time (s) ====" )
print( cDecompress )
print( "==== Compressed Size (MB) ====" )
print( cBytes / 2**20 )
print( "==== Compression Ratio (%) ====" )
print( cRatio )

print( "TODO: test decompression times" )
print( "TODO: test read/write times" )

#print( "Compression time: %f s" %(t1-t0) )
#print( "Original DM4 size: %d MB" % (originalBytes /(2**20)) )
#uncompressedBytes = np.prod( dm4image.shape )
#print( "UNcompressed size: %d MB" % (uncompressedBytes/(2**20)) )
#print( "Compressed size: %d MB" % (header['compressedBytes']/(2**20)) )
#print( "Effective compression ratio: %d %% " % (100.0 * originalBytes/header['compressedBytes']) )



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 13 10:24:09 2016

@author: rmcleod
"""
import numpy as np
import matplotlib.pyplot as plt
import blosc

SHUFFLE = blosc.BITSHUFFLE  # Wow BITSHUFFLE IS FAST
nThreadsList = np.arange( 12, 48+1, 6 )
shuffleList = np.arange(0,2+1)
compressorList = np.array( ['lz4', 'zlib', 'zstd' ] )
clevelList = np.arange(1,6+1)
doseLevelList = [ 0.1, 0.25, 0.5, 1.0, 1.5, 2.0, 4.0 ]
testRuns = 1
blockSizeList = np.array( [2**15, 2**16, 2**17, 2**18, 2**19, 2**20, 2**21, 2**22], dtype='int64' )

colorList = ['steelblue', 'forestgreen', 'purple']

optiShape = [len(doseLevelList), len(compressorList), len(clevelList), 
             len(nThreadsList), len(blockSizeList), testRuns]
            
cCompress = np.load( "cCompress.npy"  )
cBytes  = np.load( "cBytes.npy" )
cRatio = np.load( "cRatio.npy" )
cDecompress = np.load( "cDecompress.npy" )

# Average over runs
# (could take std too)
cCompress = np.mean( cCompress, axis=-1 )
cBytes = np.mean( cBytes, axis=-1 )
cRatio = np.mean( cRatio, axis=-1 )
cDecompress = np.mean( cDecompress, axis=-1 )

originalBytes = 50*4096*4096*4 # Guess of size if we saved as float-32

cCompressRate = originalBytes/4 / cCompress / 1E6
cDecompressRate = originalBytes/4 / cDecompress / 1E6

best = np.unravel_index( np.argmax( cCompressRate[3,0,...] ), cCompressRate.shape[2:] )
print( "LZ4 best compression rate: %f at clevel=%d, n_threads=%d, blocksize=%d" % \
      (np.max( cCompressRate[3,0,...] ), clevelList[best[0]], nThreadsList[best[1]], blockSizeList[best[2]]  ) )
best = np.unravel_index( np.argmax( cCompressRate[3,1,...] ), cCompressRate.shape[2:] )
print( "Zlib best compression rate: %f at clevel=%d, n_threads=%d, blocksize=%d" % \
      (np.max( cCompressRate[3,1,...] ), clevelList[best[0]], nThreadsList[best[1]], blockSizeList[best[2]]  ) )
best = np.unravel_index( np.argmax( cCompressRate[3,2,...] ), cCompressRate.shape[2:] )
print( "Zstd best compression rate: %f at clevel=%d, n_threads=%d, blocksize=%d" % \
      (np.max( cCompressRate[3,2,...] ), clevelList[best[0]], nThreadsList[best[1]], blockSizeList[best[2]]  ) )



#for I, doseLevel in enumerate( doseLevelList ):
#    for J, compressor in enumerate(compressorList):
#        for K, clevel in enumerate( clevelList ):
#            for N, N_THREADS in enumerate( nThreadsList ):
#                for B, blocksize in enumerate( blockSizeList ):
    
#### Compression rate ####
maxLz4cRate = np.max( np.max( np.max( cCompressRate[:,0,...], axis=3 ), axis=2 ), axis=1 )
maxZLibcRate = np.max( np.max( np.max( cCompressRate[:,1,...], axis=3 ), axis=2 ), axis=1 )
maxZStdcRate = np.max( np.max( np.max( cCompressRate[:,2,...], axis=3 ), axis=2 ), axis=1 )
maxLz4deRate = np.max( np.max( np.max( cDecompressRate[:,0,...], axis=3 ), axis=2 ), axis=1 )
maxZLibdeRate = np.max( np.max( np.max( cDecompressRate[:,1,...], axis=3 ), axis=2 ), axis=1 )
maxZStddeRate = np.max( np.max( np.max( cDecompressRate[:,2,...], axis=3 ), axis=2 ), axis=1 )
# Something wrong with the decompression benchmark...


uint4CRate = originalBytes/4 / 0.8 / 1E6
plt.figure()
plt.plot( doseLevelList, maxLz4cRate, '.-', 
         linewidth=1.5, markeredgecolor='k', color=colorList[0], label='lz4' )
plt.plot( doseLevelList, maxLz4deRate, '--', color=colorList[0], label='unlz4' )
plt.plot( doseLevelList, maxZLibcRate, '.-',
         linewidth=1.5, markeredgecolor='k', color=colorList[1], label='zlib' )
plt.plot( doseLevelList, maxZLibdeRate, '--', color=colorList[1], label='unzlib' )
plt.plot( doseLevelList, maxZStdcRate, '.-', 
         linewidth=1.5, markeredgecolor='k', color=colorList[2], label='zstd' )
plt.plot( doseLevelList, maxZStddeRate, '--', color=colorList[2], label='unzstd' )
plt.plot( doseLevelList, uint4CRate*np.ones_like(doseLevelList), '-', 
         linewidth=1.5, color='darkorange', label='uint4' )
plt.legend( loc='best' )
plt.xlabel( "Dose ($e^-/pix/frame$)" )
plt.ylabel( "Compression Rate ($MB/s$)" )
plt.savefig( "DoseRate_vs_CompressionRate.png" )

#### Compression ratio ####
maxLz4cRatio = np.max( np.max( np.max( cRatio[:,0,...], axis=3 ), axis=2 ), axis=1 )
maxZLibcRatio = np.max( np.max( np.max( cRatio[:,1,...], axis=3 ), axis=2 ), axis=1 )
maxZStdcRatio = np.max( np.max( np.max( cRatio[:,2,...], axis=3 ), axis=2 ), axis=1 )


plt.figure()
plt.plot( doseLevelList, maxLz4cRatio, '.-', 
         linewidth=1.5, markeredgecolor='k', color=colorList[0], label='lz4' )
plt.plot( doseLevelList, maxZLibcRatio, '.-', 
         linewidth=1.5, markeredgecolor='k', color=colorList[1], label='zlib' )
plt.plot( doseLevelList, maxZStdcRatio, '.-', 
         linewidth=1.5, markeredgecolor='k', color=colorList[2], label='zstd' )
plt.plot( doseLevelList, 800.0 * np.ones_like(doseLevelList), '-', 
         linewidth=1.5, color='darkorange', label='uint4' )
plt.legend( loc='best' )
plt.xlabel( "Dose ($e^-/pix/frame$)" )
plt.ylabel( "Compression Ratio ($\%$)" )
plt.savefig( "DoseRate_vs_CompressionRatio.png" )


# Compression Rate, Ratio, versus clevel
# Just plot for clevel 1?
for I, doseLevel in enumerate( doseLevelList ):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for J, compressor in enumerate( compressorList ):
        
        bestCompressRate = np.zeros_like( clevelList )
        optCRatio = np.zeros_like( clevelList )
        for K in np.arange( len( clevelList ) ):
            bestNB = np.unravel_index( np.argmax( cCompressRate[I,J,K,...] ) , cCompressRate.shape[-2:] )
            bestCompressRate[K] = cCompressRate[I,J,K,bestNB[0],bestNB[1]]
            optCRatio[K] = cRatio[I,J,K,bestNB[0], bestNB[1]]
        plt.semilogx( bestCompressRate, optCRatio, '.-', color=colorList[J], label=compressorList[J],
                 linewidth=1.5, markeredgecolor='k' )
        ax.annotate( '1', xy=(bestCompressRate[0],optCRatio[0]) )
    plt.legend( loc='best' )
    plt.xlabel( "Compression Rate ($MB/s$)" )
    plt.ylabel( "Compression Ratio ($\%$)" )
    plt.title( "Dose level $%.1f e^-/pix/frame$" % doseLevelList[I] )
    plt.savefig( "clevelPlot_dose%f.png" % doseLevelList[I] )
    
    
#### FIND BEST COMPRESSION LEVEL, NTHREADS, BLOCKSIZE FOR COMPRESSIONRATE AND CRATIO             

# I think scaling by the number of threads is a little too confusing?
fig = plt.figure()
I = 3 # 1.0 e-/pix
K = 0
B = 1
for J, compressor in enumerate( compressorList ):
    plt.plot( nThreadsList, cCompressRate[I,J,K,:,B], '.-', color=colorList[J], label=compressorList[J],
             linewidth=1.5, markeredgecolor='k' )
plt.title( "Thread scaling for $1.0 e^-/pix/frame$, clevel $1$, block size $64k$" )
plt.xlabel( "Number of threads" )
plt.ylabel( "Compression Rate ($MB/s$)" )
plt.legend( loc='best' )
plt.savefig( "CompressionRate_threadScaling.png" )


fig = plt.figure()
I = 3 # 1.0 e-/pix
K = 0
B = 1
for J, compressor in enumerate( compressorList ):
    normCompressRate = cCompressRate[I,J,K,:,B] / np.array( nThreadsList )
    plt.plot( nThreadsList, normCompressRate, '.-', color=colorList[J], label=compressorList[J],
             linewidth=1.5, markeredgecolor='k' )
plt.title( "Thread scaling for $1.0 e^-/pix/frame$, clevel $1$, block size $64k$" )
plt.xlabel( "Number of threads" )
plt.ylabel( "Normalized Compression Rate ($MB/s/thread$)" )
plt.legend( loc='best' )
plt.savefig( "NormedCompressionRate_threadScaling.png" )

fig, ax1 = plt.subplots( figsize=(10,8) )
I = 3 # 1.0 e-/pix
K = 0
N = 5
for J, compressor in enumerate( compressorList ):
    ax1.plot( blockSizeList/(2**10), cRatio[I,J,K,N,:], '+-', color=colorList[J], label="Ratio " + compressorList[J],
             linewidth=1.5, markeredgecolor='k' )
ax1.set_xlabel( "Block size ($kB$)" )
ax1.set_ylabel( "Compression Ratio ($\%$)" )
ax1.legend( loc='center' )
ax2 = ax1.twinx()
for J, compressor in enumerate( compressorList ):
    ax2.plot( blockSizeList/(2**10), cCompressRate[I,J,K,N,:], 'o--', color=colorList[J], label="Rate " + compressorList[J],
             linewidth=1.5, markeredgecolor='k' )
ax2.set_ylabel( "Compression Rate ($MB/s$)" )
plt.title( "Scaling with block size" )
ax2.legend( loc='right' )
plt.savefig( "blockScaling.png" )



#plt.figure()
#plt.plot( doseLevelList, compressStackSize, 'k.-'  )
#plt.xlabel( r"Dose Rate $(e^-/\AA^2)$" )
#plt.ylabel( r"Compressed file size (GB)" )
#plt.title( r"Expected stack size for a total dose of $80 e^-/\AA^2 (zstd5)$" )
#plt.savefig( "ExpectedMRCZSizeFor80e.png" )


import glob, sys, os, os.path as path
import numpy as np
import mrcz
from distutils.version import StrictVersion
import shutil

args = sys.argv
directory = '.'
"""
This script is used for updating a directory of older MRCZ files from versions
<= 0.4.1 to the 0.5.0 standard, so that the volume sampling is set to the 
default (Mx, My, Mz) = [0, 0, 1].
"""

if len(sys.argv) > 1:
    directory = sys.argv[1]

mrcz_files = glob.glob(path.join(directory, '*.mrc'), recursive=False)
mrcz_files.extend(glob.glob(path.join(directory, '*.mrcz'), recursive=False))

backup_dir = path.join(directory, 'backup')
os.makedirs(backup_dir)
for filename in mrcz_files:
    shutil.copy(filename, path.join(backup_dir, path.basename(filename)))

    data, meta = mrcz.readMRC(filename)

    compressor = meta['compressor'] if 'compressor' in meta else None
    clevel = meta['clevel'] if 'clevel' in meta else 1
    voltage = meta['voltage'] if 'voltage' in meta else 0.0
    C3 = meta['C3'] if 'C3' in meta else 0.0
    gain = meta['gain'] if 'gain' in meta else 1.0
    pixelsize = meta['pixelsize'] if 'pixelsize' in meta else [0.1,0.1,0.1]
    pixelunits = meta['pixelunits'] if 'pixelunits' in meta else u'\\AA'
    
    if 'packedBytes' in meta:
        meta.pop('packedBytes')
    if 'minImage' in meta:
        meta.pop('minImage')
    if 'maxImage' in meta:
        meta.pop('maxImage')
    if 'meanImage' in meta:
        meta.pop('meanImage')
    if 'extendedBytes' in meta:
        meta.pop('extendedBytes')

    mrcz.writeMRC(data, filename, meta=meta,
        pixelsize=pixelsize, pixelunits=pixelunits,
        voltage=voltage, C3=C3, gain=gain,
        compressor=compressor, clevel=clevel)
    
    # Try and read every file back in.
    try:
        mrcz.readMRC(filename)
    except:
        raise ValueError('{} could not be re-read'.format(filename))
    
print('__All__ originals were backed up to %s, remember to delete them' % path.join(directory, 'backup'))

    